import jax

# noinspection PyUnresolvedReferences
import lab.jax

from .minimise import *

# noinspection PyUnresolvedReferences
from .. import *

# We will need `float64`s
jax.config.update("jax_enable_x64", True)



import logging

import lab.autograd as B
from autograd import grad as only_grad, value_and_grad
from plum import convert

from ..minimise import make_l_bfgs_b, make_adam, exception

__all__ = ["minimise_l_bfgs_b", "minimise_adam"]

log = logging.getLogger(__name__)


def _wrap_f(vs, names, f, jit, _convert):
    if jit:
        raise ValueError("There is no JIT for AutoGrad.")

    # Differentiable assignments will overwrite the variables, so make a copy.
    vs_copy = vs.copy()

    # Keep track of function evaluations.
    f_evals = []

    def f_vectorised(x, *args):
        vs_copy.set_latent_vector(x, *names, differentiable=True)
        out = convert(f(vs_copy, *args), tuple)
        return out[0], out[1:]

    def f_wrapped(x, *args):
        x = B.cast(vs.dtype, x)

        # Compute objective function value and gradient.
        try:
            if args == ():
                # Don't need to update `args`.
                obj_value, grad = value_and_grad(lambda x_: f_vectorised(x_)[0])(x)
            else:
                # AutoGrad doesn't allow us to only compute the gradient with respect
                # to the first output, so we have to do it this way, incurring an
                # extra forward passs...
                obj_value, new_args = f_vectorised(x, *args)
                grad = only_grad(lambda x_: f_vectorised(x_, *args)[0])(x)
                args = new_args

        except Exception as e:
            return exception(x, args, e)

        # Perform requested conversion.
        obj_value, grad = _convert(obj_value, grad)

        f_evals.append(obj_value)
        return (obj_value, args), grad

    return f_evals, f_wrapped


minimise_l_bfgs_b = make_l_bfgs_b(_wrap_f)
minimise_adam = make_adam(_wrap_f)



import importlib
import logging
import re
from itertools import product
from typing import Union
from functools import reduce
from operator import mul

import lab as B
from plum import Dispatcher

__all__ = ["lazy_tf", "lazy_torch", "lazy_jnp", "pack", "unpack", "match"]

_dispatch = Dispatcher()

log = logging.getLogger(__name__)


class LazyModule:
    """A module that loads once an attribute is requested.

    Args:
        name (str): Name of module.
    """

    def __init__(self, name):
        self.name = name
        self.module = None

    def __getattr__(self, item):
        # Load module.
        if self.module is None:
            self.module = importlib.import_module(self.name)

        return getattr(self.module, item)


lazy_tf = LazyModule("tensorflow")
lazy_torch = LazyModule("torch")
lazy_jnp = LazyModule("jax.numpy")


@_dispatch
def pack(*objs: B.Numeric):
    """Pack objects.

    Args:
        *objs (tensor): Objects to pack.

    Returns:
        tensor: Vector representation of the objects.
    """
    return B.concat(*[B.flatten(obj) for obj in objs], axis=0)


@_dispatch
def unpack(package: B.Numeric, *shapes):
    """Unpack vector.

    Args:
        package (tensor): Tensor to unpack.
        *shapes (shape): Shapes of objects to unpack.

    Returns:
        list[tensor]: Original objects.
    """
    if B.rank(package) != 1:
        raise ValueError("Package must be a vector.")

    # Unpack package.
    lengths = [reduce(mul, shape, 1) for shape in shapes]
    i, outs = 0, []
    for length, shape in zip(lengths, shapes):
        outs.append(B.reshape(package[i : i + length], *shape))
        i += length
    return outs


def match(pattern, target):
    """Match a string exactly against a pattern, where the pattern is a regex
    with '*' as the only active character.

    Args:
        pattern (str): Pattern.
        target (str): Target.

    bool: `True` if `pattern` matches `target`.
    """
    pattern = "".join(".*" if c == "*" else re.escape(c) for c in pattern)
    return bool(re.match("^" + pattern + "$", target))



import jax.numpy as jnp
import lab as B
import numpy as np
import pytest
import tensorflow as tf
import torch
import varz.autograd
import varz.jax
import varz.tensorflow
import varz.torch
from varz import Vars
from varz.minimise import _convert_and_validate_names

from .util import approx, Value, OutStream

_rate = 5e-2
_minimise_method_params = []
for dtype_name in ["float32", "float64"]:
    _minimise_method_params.extend(
        [
            (getattr(np, dtype_name), varz.minimise_l_bfgs_b, {}),
            (
                getattr(np, dtype_name),
                varz.minimise_adam,
                {"rate": _rate, "local_rates": False},
            ),
            (
                getattr(np, dtype_name),
                varz.minimise_adam,
                {"rate": _rate, "local_rates": True},
            ),
        ]
    )
    for backend in [tf, torch, jnp]:
        _minimise_method_params.extend(
            [
                (
                    getattr(backend, dtype_name),
                    varz.minimise_l_bfgs_b,
                    {"jit": True},
                ),
                (
                    getattr(backend, dtype_name),
                    varz.minimise_adam,
                    {"rate": _rate, "jit": True, "local_rates": False},
                ),
                (
                    getattr(backend, dtype_name),
                    varz.minimise_adam,
                    {"rate": _rate, "jit": True, "local_rates": True},
                ),
            ]
        )


@pytest.fixture(params=_minimise_method_params)
def minimise_method(request):
    yield request.param


def test_convert_and_validate_names():
    assert _convert_and_validate_names(None) == []
    assert _convert_and_validate_names([]) == []
    assert _convert_and_validate_names("test") == ["test"]
    assert _convert_and_validate_names(["a", "b"]) == ["a", "b"]
    with pytest.raises(ValueError):
        _convert_and_validate_names(["a", ["b"]])
    with pytest.raises(ValueError):
        _convert_and_validate_names([1, "a"])


def test_docstring(minimise_method):
    assert minimise_method.__doc__ is not None


def test_no_jit_autograd():
    vs = Vars(dtype=np.float64)

    def f(vs_):
        return vs_.ubnd(name="x")

    with pytest.raises(ValueError):
        varz.autograd.minimise_l_bfgs_b(f, vs, jit=True)
    with pytest.raises(ValueError):
        varz.autograd.minimise_adam(f, vs, jit=True)


def test_minimise(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    # Initialise a variable that is not used.
    vs.ubnd(name="other")

    # Define some objective.
    def f(vs_):
        return (-3 - vs_.pos(name="x", init=1.0)) ** 2

    # Minimise it, until convergence.
    val_opt = minimise(f, vs, iters=2000, **kw_args)

    # Check for equality up to two digits.
    approx(val_opt, 9, atol=1e-2)
    approx(vs["x"], 0, atol=1e-2)


def test_minimise_auxilary_argument(minimise_method):
    # Perform the above test, but also pass an auxilary argument around.
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    # Again, initialise a variable that is not used.
    vs.ubnd(name="other")

    # Again, define some objective.
    def f(vs_, state):
        return (-3 - vs_.pos(name="x", init=1.0)) ** 2, state + 1

    # Minimise it, until convergence, but now also get the final state.
    val_opt, final_state = minimise(f, (vs, B.cast(dtype, 1)), iters=2000, **kw_args)

    # Check for equality up to two digits.
    approx(val_opt, 9, atol=1e-2)
    approx(vs["x"], 0, atol=1e-2)

    # Check that the internal state was passed around.
    assert final_state > 5


def test_minimise_disconnected_gradient(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)
    vs.ubnd(name="x")

    # Check that optimiser runs for objective that returns the constant zero.
    minimise(lambda v: B.cast(v.dtype, 0), vs, **kw_args)


def test_minimise_exception(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    # Don't use the JIT in this test, because the tracing will fail due to the
    # randomness.
    kw_args = dict(kw_args)  # Copy to prevent mutation.
    kw_args["jit"] = False
    kw_args["iters"] = 100

    first_call = Value(True)

    # Define an objective that sometimes fails after the first call.
    def f(vs_):
        if first_call.val:
            first_call.val = False
        else:
            if np.random.rand() > 0.5:
                raise Exception("Fail!")
        return vs_.ubnd(name="x", init=5.0) ** 2

    # Check that the optimiser runs.
    minimise(f, vs, **kw_args)


def test_minimise_zero_calls(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    calls = Value(0)

    def f(vs_):
        calls.val += 1
        return vs_.ubnd(name="x", init=5.0) ** 2

    # Check that running the optimiser for zero iterations only incurs a
    # single call.
    minimise(f, vs, iters=0, **kw_args)
    assert calls.val == 1


def test_minimise_trace(minimise_method):
    dtype, minimise, kw_args = minimise_method

    def f(vs_):
        return vs_.ubnd(name="x") ** 2

    # Test that `trace = False` prints nothing.
    with OutStream() as stream:
        minimise(f, Vars(dtype=dtype), trace=False, **kw_args)
        assert stream.output == ""

    # Test that `trace = False` prints something.
    with OutStream() as stream:
        minimise(f, Vars(dtype=dtype), trace=True, **kw_args)
        assert stream.output != ""



import traceback
from functools import wraps
import importlib

from .adam import ADAM

import lab as B
import numpy as np
import wbml.out as out
from plum import Dispatcher
from plum import convert, List
from scipy.optimize import fmin_l_bfgs_b

__all__ = ["minimise_l_bfgs_b", "minimise_adam"]

_dispatch = Dispatcher()


@_dispatch
def _get_minimise_f(dtype: B.NPDType, name: str):
    mod = importlib.import_module("varz.autograd.minimise")
    return getattr(mod, name)


@_dispatch
def _get_minimise_f(dtype: B.TFDType, name: str):
    mod = importlib.import_module("varz.tensorflow.minimise")
    return getattr(mod, name)


@_dispatch
def _get_minimise_f(dtype: B.TorchDType, name: str):
    mod = importlib.import_module("varz.torch.minimise")
    return getattr(mod, name)


@_dispatch
def _get_minimise_f(dtype: B.JAXDType, name: str):
    mod = importlib.import_module("varz.jax.minimise")
    return getattr(mod, name)


def _convert_and_validate_names(names):
    if names is None:
        names = []
    if isinstance(names, str):
        names = [names]
    if not isinstance(names, List[str]):
        raise ValueError("Keyword `names` must be a list of strings.")
    return names


def minimise_l_bfgs_b(
    f, vs, f_calls=10000, iters=1000, trace=False, names=None, jit=False
):  # pragma: no cover
    """Minimise a function with L-BFGS-B.

    Args:
        f (function): Function to optimise.
        vs (:class:`.vars.Vars`): Variable manager.
        f_calls (int, optional): Maximum number of function calls. Defaults to `10000`.
        iters (int, optional): Maximum number of iterations. Defaults to `1000`.
        trace (bool, optional): Show trace of optimisation. Defaults to `False`.
        names (list, optional): List of names of variables to optimise.
            Defaults to all variables.
        jit (bool, optional): Use a JIT if one is available. Defaults to `False`.

    Returns:
        float: Final objective function value.
    """
    minimise_f = _get_minimise_f(convert(vs, tuple)[0].dtype, "minimise_l_bfgs_b")
    return minimise_f(
        f=f,
        vs=vs,
        f_calls=f_calls,
        iters=iters,
        trace=trace,
        names=names,
        jit=jit,
    )


def minimise_adam(
    f,
    vs,
    iters=1000,
    rate=1e-3,
    beta1=0.9,
    beta2=0.999,
    epsilon=1e-8,
    local_rates=True,
    trace=False,
    names=None,
    jit=False,
):  # pragma: no cover
    """Minimise a function with Adam.

    Args:
        f (function): Function to optimise.
        vs (:class:`.vars.Vars`): Variable manager.
        iters (int, optional): Maximum number of iterations. Defaults to `1000`.
        rate (float, optional): Learning rate. Defaults to `1e-3`.
        beta1 (float, optional): Exponential decay for mean. Defaults to `0.9`.
        beta2 (float, optional): Exponential decay for second moment. Defaults to
            `0.999`.
        epsilon (float, optional): Small value to prevent division by zero.
            Defaults to `1e-8`.
        local_rates (bool, optional): Use local learning rates. Set to `False` to
            use one global learning rate. Defaults to `True`.
        trace (bool, optional): Show trace of optimisation. Defaults to `False`.
        names (list, optional): List of names of variables to optimise.
            Defaults to all variables.
        jit (bool, optional): Use a JIT if one is available. Defaults to `False`.

    Returns:
        float: Final objective function value.
    """
    minimise_f = _get_minimise_f(convert(vs, tuple)[0].dtype, "minimise_adam")
    return minimise_f(
        f=f,
        vs=vs,
        iters=iters,
        rate=rate,
        beta1=beta1,
        beta2=beta2,
        epsilon=epsilon,
        local_rates=local_rates,
        trace=trace,
        names=names,
        jit=jit,
    )


def make_l_bfgs_b(wrap_f):
    """Create `minimise_l_bfgs_b` given a function wrapper.

    Args:
        wrap_f (function): Function wrapper.

    Returns:
        function: `minimise_l_bfgs_b`.
    """

    @wraps(minimise_l_bfgs_b)
    def _minimise_l_bfgs_b(
        f, vs, f_calls=10000, iters=1000, trace=False, names=None, jit=False
    ):
        names = _convert_and_validate_names(names)

        # Extract variable container and auxilary arguments from argument specification.
        vs = convert(vs, tuple)
        vs, args = vs[0], vs[1:]

        # Run function once to ensure that all variables are initialised and
        # available.
        res = convert(f(vs, *args), tuple)
        val_init, args = res[0], res[1:]

        # SciPy doesn't perform zero iterations, so handle that edge case
        # manually.
        if iters == 0 or f_calls == 0:
            return B.squeeze((B.to_numpy(val_init),) + args)

        # Extract initial value.
        x0 = B.to_numpy(vs.get_latent_vector(*names))

        # The optimiser expects to get `float64`s.
        def _convert(*xs):
            return [B.cast(np.float64, B.to_numpy(x)) for x in xs]

        # Wrap the function and get the list of function evaluations.
        f_vals, f_wrapped = wrap_f(vs, names, f, jit, _convert)

        # Maintain a state for the auxilary arguments.
        state = {"args": args}

        # Wrap `f_wrapped` to take the auxilary from the global state and then update
        # the global state.

        def f_wrapped_self_passing(x):
            (obj_value, state["args"]), grad = f_wrapped(x, *state["args"])
            return obj_value, grad

        # Perform optimisation routine.
        def perform_minimisation(callback_=lambda _: None):
            return fmin_l_bfgs_b(
                func=f_wrapped_self_passing,
                x0=x0,
                maxiter=iters,
                maxfun=f_calls,
                callback=callback_,
                disp=0,
            )

        if trace:
            # Print progress during minimisation.
            with out.Progress(
                name='Minimisation of "{}"'.format(f.__name__), total=iters
            ) as progress:

                def callback(_):
                    progress({"Objective value": np.min(f_vals)})

                x_opt, val_opt, info = perform_minimisation(callback)

            with out.Section("Termination message"):
                out.out(convert(info["task"], str))
        else:
            # Don't print progress; simply perform minimisation.
            x_opt, val_opt, info = perform_minimisation()

        vs.set_latent_vector(x_opt, *names)  # Assign optimum.
        args = state["args"]  # Get auxilary arguments at final state.

        return B.squeeze((val_opt,) + args)  # Return optimal value.

    return _minimise_l_bfgs_b


def make_adam(wrap_f):
    """Create `minimise_adam` given a function wrapper.

    Args:
        wrap_f (function): Function wrapper.

    Returns:
        function: `minimise_adam`.
    """

    @wraps(minimise_adam)
    def _minimise_adam(
        f,
        vs,
        iters=1000,
        rate=1e-3,
        beta1=0.9,
        beta2=0.999,
        epsilon=1e-8,
        local_rates=True,
        trace=False,
        names=None,
        jit=False,
    ):
        names = _convert_and_validate_names(names)

        # Extract variable container and auxilary arguments from argument specification.
        vs = convert(vs, tuple)
        vs, args = vs[0], vs[1:]

        # Run function once to ensure that all variables are initialised and
        # available.
        res = convert(f(vs, *args), tuple)
        val_init, args = res[0], res[1:]

        # Handle the edge case of zero iterations.
        if iters == 0:
            return B.squeeze((B.to_numpy(val_init),) + args)

        # Extract initial value.
        x0 = B.to_numpy(vs.get_latent_vector(*names))

        # Wrap the function.
        _, f_wrapped = wrap_f(vs, names, f, jit, B.to_numpy)

        # Maintain a state for the auxilary arguments.
        state = {"args": args}

        def perform_minimisation(callback_=lambda _: None):
            # Perform optimisation routine.
            x = x0
            obj_value = None
            adam = ADAM(
                rate=rate,
                beta1=beta1,
                beta2=beta2,
                epsilon=epsilon,
                local_rates=local_rates,
            )

            for i in range(iters):
                (obj_value, state["args"]), grad = f_wrapped(x, *state["args"])
                callback_(obj_value)
                x = adam.step(x, grad)

            return x, obj_value

        if trace:
            # Print progress during minimisation.
            with out.Progress(
                name='Minimisation of "{}"'.format(f.__name__), total=iters
            ) as progress:

                def callback(obj_value):
                    progress({"Objective value": obj_value})

                x_opt, obj_value = perform_minimisation(callback)
        else:
            x_opt, obj_value = perform_minimisation()

        vs.set_latent_vector(x_opt, *names)  # Assign optimum.
        args = state["args"]  # Get auxilary arguments at final state.

        return B.squeeze((obj_value,) + args)  # Return last objective value.

    return _minimise_adam


def exception(x, args, e):
    """In the case that an exception is raised during function evaluation,
    print a warning and return NaN for the function value and gradient.

    Args:
        x (tensor): Current input.
        args (tuple): Current auxilary arguments.
        e (:class:`Exception`): Caught exception.

    Returns:
        tuple: Return value containing NaN for the objective value and NaNs for the
            gradient.
    """
    with out.Section("Caught exception during function evaluation"):
        out.out(traceback.format_exc().strip())
    grad_nan = np.empty(x.shape)
    grad_nan[:] = np.nan
    return (np.nan, args), grad_nan



import abc
import inspect
from abc import ABCMeta, abstractmethod
from functools import wraps
from inspect import isfunction
from itertools import repeat

from .vars import Provider

__all__ = [
    "sequential",
    "namespace",
    "Struct",
    "Unbounded",
    "Positive",
    "Bounded",
    "LowerTriangular",
    "PositiveDefinite",
    "Orthogonal",
    "parametrised",
]


class _RedirectedProvided(Provider):
    """Redirect all methods of :class:`.vars.Provider` to a method `_get_var`.

    Args:
        vs (:class:`.vs.Vars`): Variable container to wrap.
    """

    def __init__(self, vs):
        self._vs = vs

    @abc.abstractmethod
    def _get_var(self, getter, args, kw_args):  # pragma: no cover
        pass

    def unbounded(self, *args, **kw_args):
        return self._get_var(self._vs.unbounded, args, kw_args)

    def positive(self, *args, **kw_args):
        return self._get_var(self._vs.positive, args, kw_args)

    def bounded(self, *args, **kw_args):
        return self._get_var(self._vs.bounded, args, kw_args)

    def lower_triangular(self, *args, **kw_args):
        return self._get_var(self._vs.lower_triangular, args, kw_args)

    def positive_definite(self, *args, **kw_args):
        return self._get_var(self._vs.positive_definite, args, kw_args)

    def orthogonal(self, *args, **kw_args):
        return self._get_var(self._vs.orthogonal, args, kw_args)

    def __getitem__(self, name):
        return self._vs[name]


class Sequential(_RedirectedProvided):
    """A variable provider that wraps a :class:`.vars.Vars` object and
    automatically names unnamed variables sequentially (0, 1, 2, ...) with a
    possible prefix which defaults to "var".

    Args:
        vs (:class:`.vs.Vars`): Variable container to wrap.
        prefix (str): Prefix for names in the sequence.
    """

    def __init__(self, vs, prefix):
        _RedirectedProvided.__init__(self, vs)
        self._prefix = prefix
        self._count = 0

    def _get_var(self, getter, args, kw_args):
        if "name" not in kw_args:
            kw_args["name"] = f"{self._prefix}{self._count}"
            self._count += 1
        return getter(*args, **kw_args)


def _to_sequential(x, prefix):
    """Convert a variable provider to sequential, if it is one.

    Args:
        x (object): Object to convert.
        prefix (str): Prefix for names in the sequence.

    Returns:
        object: `x` converted to sequential if it is a variable provider,
            otherwise just `x`.
    """
    return Sequential(x, prefix) if isinstance(x, Provider) else x


def _extract_prefix_and_f(prefix_or_f, default):
    """Extract the prefix and function.

    Args:
        prefix_or_f (object): Either a prefix or a function.
        default (str): Default prefix.

    Returns:
        tuple: Tuple containing the predict and the function. If a function
            was given, the prefix defaults to an empty string; and if a prefix
            was given, the function defaults to `None`.
    """
    if isfunction(prefix_or_f):
        # Used as a decorator.
        return default, prefix_or_f
    else:
        # Not used as decorator and prefix is specified.
        return prefix_or_f, None


def sequential(prefix="var"):
    """Decorator that generates variable names for unnamed variables
    sequentially.

    Args:
        prefix (str, optional): Prefix to prepend to the name of generated
            variables. Defaults to "var".
    """
    prefix, f = _extract_prefix_and_f(prefix, default="var")

    def decorator(f_):
        @wraps(f_)
        def wrapped_f(*args, **kw_args):
            # Replace all variable containers with their sequential variants.
            args = [_to_sequential(x, prefix) for x in args]
            kw_args = {k: _to_sequential(v, prefix) for k, v in kw_args.items()}

            return f_(*args, **kw_args)

        return wrapped_f

    # Return the decorated function directly if a function was given.
    return decorator if f is None else decorator(f)


class Namespace(_RedirectedProvided):
    """A variable provider that wraps a :class:`.vars.Vars` object and automatically
    prepends a prefix to named variables.

    Args:
        vs (:class:`.vs.Vars`): Variable container to wrap.
        prefix (str): Prefix for names.
    """

    def __init__(self, vs, prefix):
        _RedirectedProvided.__init__(self, vs)
        # Ensure that the prefix ends with a ".".
        if not prefix.endswith("."):
            prefix = prefix + "."
        self._prefix = prefix

    def _get_var(self, getter, args, kw_args):
        if "name" in kw_args:
            kw_args["name"] = self._prefix + kw_args["name"]
        return getter(*args, **kw_args)


def _to_namespace(x, prefix):
    """Convert a variable provider to a namespace, if it is one.

    Args:
        x (object): Object to convert.
        prefix (str): Prefix for names.

    Returns:
        object: `x` converted to a namespace if it is a variable provider,
            otherwise just `x`.
    """
    return Namespace(x, prefix) if isinstance(x, Provider) else x


def namespace(prefix):
    """Decorator that prefixes named variables.

    Args:
        prefix (str): Prefix to prepend to the name of named variables.
    """

    def decorator(f_):
        @wraps(f_)
        def wrapped_f(*args, **kw_args):
            # Replace all variable containers with their namespace variants.
            args = [_to_namespace(x, prefix) for x in args]
            kw_args = {k: _to_namespace(v, prefix) for k, v in kw_args.items()}

            return f_(*args, **kw_args)

        return wrapped_f

    return decorator


class Struct(_RedirectedProvided):
    """A variable provider that wraps a :class:`.vars.Vars` object and allows variables
    to be automatically named by getting attributes and indexing.

    Args:
        vs (:class:`.vs.Vars`): Variable container to wrap.
        path (str, optional): Path. Defaults to no path.
    """

    def __init__(self, vs, path=None):
        _RedirectedProvided.__init__(self, vs)
        self._path = path

    def _resolve_path(self, key, separator=""):
        if self._path:
            return f"{self._path}{separator}{key}"
        else:
            return str(key)

    def _get_var(self, getter, args, kw_args):
        if self._path:
            if "name" in kw_args:
                name = f'{self._path}.{kw_args["name"]}'
            else:
                name = self._path
            kw_args["name"] = name
        return getter(*args, **kw_args)

    def __getattr__(self, item):
        path = self._resolve_path(item, separator=".")
        return Struct(self._vs, path)

    def __getitem__(self, item):
        if isinstance(item, int) and item < 0:
            item += len(self)
        path = self._resolve_path(f"[{item}]")
        return Struct(self._vs, path)

    def __iter__(self):
        state = {"counter": -1}

        def get_next():
            state["counter"] += 1
            return self[state["counter"]]

        return (get_next() for _ in repeat(True))

    def __call__(self):
        return self._vs[self._path]

    def all(self):
        """Get a regex that matches everything in the current path.

        Returns:
            str: Regex.
        """
        return self._resolve_path("*", separator=".")

    def __len__(self):
        i = 0
        while any(
            name.startswith(self._resolve_path(f"[{i}]")) for name in self._vs.names
        ):
            i += 1
        return i

    def __next__(self):
        return self[len(self)]

    def up(self, level=None):
        """Go up one level.

        Args:
            level (str, optional): Assert that the name of the level to go up is equal
                to `level`. Does not perform this assertion if `level` is not given.

        Returns:
            :class:`.spec.Struct`: Struct.
        """
        parts = self._path.split(".")
        if level is not None and parts[-1] != level:
            raise AssertionError(
                f'Cannot go up level "{level}" because the current path is "{self._path}".'
            )
        return Struct(self._vs, ".".join(parts[:-1]))

    def assign(self, value):
        """Assign a value.

        Args:
            value (tensor): Value to assign.
        """
        self._vs.assign(self._path, value)


class VariableType(metaclass=ABCMeta):
    """A type of a variable. Any arguments are passed to the appropriate method
    of :class:`.vars.Vars` to instantiate the variable."""

    def __init__(self, *args, **kw_args):
        self.args = args
        self.kw_args = kw_args

    def _get_var(self, method, name, init):
        # Copy dict to not modify the original.
        kw_args = dict(self.kw_args)

        # Set initial value.
        if init is not None and "init" in kw_args and kw_args["init"] is not None:
            raise ValueError(
                f'Initial value doubly specified: "{init}" and "{kw_args["init"]}".'
            )
        kw_args["init"] = init

        # Set name.
        if "name" in kw_args and kw_args["name"] is not None:
            raise ValueError(f'Name doubly specified: {name} and {kw_args["name"]}.')
        kw_args["name"] = name

        return method(*self.args, **kw_args)

    @abstractmethod
    def instantiate(self, vs, name, init):  # pragma: no cover
        """Instantiate the variable.

        Args:
            vs (:class:`.vars.Vars`): Variable container to extract the variable
                from.
            name (str): Name of the variable.
            init (object): Initial value.
        """


class Unbounded(VariableType):
    """Type of an unbounded variable."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.unbounded, name, init)


class Positive(VariableType):
    """Type of a positive variable."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.positive, name, init)


class Bounded(VariableType):
    """Type of a bounded variable."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.bounded, name, init)


class LowerTriangular(VariableType):
    """Type of a lower-triangular matrix."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.lower_triangular, name, init)


class PositiveDefinite(VariableType):
    """Type of a positive-definite matrix."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.positive_definite, name, init)


class Orthogonal(VariableType):
    """Type of an orthogonal matrix."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.orthogonal, name, init)


def parametrised(prefix=""):
    """Decorator to specify variables with types.

    Args:
        prefix (str, optional): Prefix to prepend to the name of generated
            variables. Defaults to no prefix.
    """

    prefix, f = _extract_prefix_and_f(prefix, default="")

    def decorator(f_):
        @wraps(f_)
        def wrapped_f(*args, **kw_args):
            signature = inspect.signature(f_)

            # Convert all arguments to keyword arguments, and store them here.
            filled_kwargs = {}

            # Look for variable container.
            values = args + tuple(kw_args.values())
            num_containers = sum([isinstance(x, Provider) for x in values])
            if num_containers == 0:
                raise ValueError("No variable container found.")
            elif num_containers > 1:
                raise ValueError("Multiple variable containers found.")
            else:
                # There is exactly only variable container. Find it.
                vs = [x for x in values if isinstance(x, Provider)][0]

            # Walk through the arguments.
            for name, parameter in signature.parameters.items():
                annotation = parameter.annotation

                # Instantiate uninstantiated variable types.
                if isinstance(annotation, type) and issubclass(
                    annotation, VariableType
                ):
                    annotation = annotation()

                if isinstance(annotation, VariableType):
                    # Parameter is a variable that needs to be extracted.

                    # Check for default value.
                    if parameter.default is not parameter.empty:
                        init = parameter.default
                    else:
                        init = None

                    # Store the instantiated variable.
                    filled_kwargs[name] = annotation.instantiate(
                        vs, prefix + name, init
                    )

                else:
                    # Parameter is a regular parameter. Find it.

                    if len(args) > 0:
                        # Positional arguments left. Must be next one.
                        filled_kwargs[name] = args[0]
                        args = args[1:]
                    else:
                        # No position arguments left. Extract from keywords.
                        try:
                            filled_kwargs[name] = kw_args[name]
                            del kw_args[name]
                        except KeyError:
                            # Also not found in keywords. Resort to default
                            # value. If a default value is not given,
                            # the user did not specify a positional argument.
                            if parameter.default is parameter.empty:
                                raise ValueError(
                                    f'Positional argument "{name}" not given.'
                                )
                            else:
                                filled_kwargs[name] = parameter.default

            # Ensure that everything is parsed.
            if len(args) > 0:
                raise ValueError(f"{len(args)} positional argument(s) not parsed.")
            if len(kw_args) > 0:
                raise ValueError(
                    f"{len(kw_args)} keyword argument(s) not "
                    f'parsed: {", ".join(kw_args.keys())}.'
                )

            return f_(**filled_kwargs)

        return wrapped_f

    # Return the decorated function directly if a function was given.
    return decorator if f is None else decorator(f)



import logging

import lab.jax as B
import numpy as np
from jax import value_and_grad
from plum import convert

from ..minimise import make_l_bfgs_b, make_adam, exception

__all__ = ["minimise_l_bfgs_b", "minimise_adam"]

log = logging.getLogger(__name__)


def _wrap_f(vs, names, f, jit, _convert):
    # Differentiable assignments will overwrite the variables, so make a copy.
    vs_copy = vs.copy()

    # Keep track of function evaluations.
    f_evals = []

    def f_vectorised(x, *args):
        vs_copy.set_latent_vector(x, *names, differentiable=True)
        out = convert(f(vs_copy, *args), tuple)
        return out[0], out[1:]

    f_value_and_grad = value_and_grad(f_vectorised, has_aux=True)

    if jit:
        f_value_and_grad = B.jit(f_value_and_grad)

    def f_wrapped(x, *args):
        x = B.cast(vs.dtype, x)

        # Compute objective function value and gradient.
        try:
            (obj_value, args), grad = f_value_and_grad(x, *args)
        except Exception as e:
            return exception(x, args, e)

        # Perform requested conversion.
        obj_value, grad = _convert(obj_value, grad)

        # The gradient may not have the right memory layout, which sometimes cannot
        # be adjusted. We therefore make a copy, which can always be freely manipulated.
        grad = np.array(grad)

        f_evals.append(obj_value)
        return (obj_value, args), grad

    return f_evals, f_wrapped


minimise_l_bfgs_b = make_l_bfgs_b(_wrap_f)
minimise_adam = make_adam(_wrap_f)



import lab as B

__all__ = ["ADAM"]


class ADAM:
    """Bare-bones implementation of ADAM.

    Args:
        rate (float, optional): Learning rate. Defaults to `1e-3`.
        beta1 (float, optional): Exponential decay for mean. Defaults to `0.9`.
        beta2 (float, optional): Exponential decay for second moment. Defaults to
            `0.999`.
        epsilon (float, optional): Small value to prevent division by zero.
            Defaults to `1e-8`.
        local_rates (bool, optional): Use local learning rates. Set to `False` to
            use one global learning rate. Defaults to `True`.
    """

    def __init__(
        self, rate=1e-3, beta1=0.9, beta2=0.999, epsilon=1e-8, local_rates=False
    ):
        self.rate = rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.local_rates = local_rates

        self.m = None
        self.v = None
        self.i = 0

    def step(self, x, grad):
        """Perform a gradient step.

        Args:
            x (tensor): Current input value. This value will be updated in-place.
            grad (tensor): Current gradient.

        Returns:
            tensor: `x` after updating `x` in-place.
        """
        if self.m is None or self.v is None:
            self.m = B.zeros(x)
            self.v = B.zeros(x)

        # Update estimates of moments.
        self.m *= self.beta1
        self.m += (1 - self.beta1) * grad
        self.v *= self.beta2
        self.v += (1 - self.beta2) * grad ** 2

        # Correct for bias of initialisation.
        m_corr = self.m / (1 - self.beta1 ** (self.i + 1))
        v_corr = self.v / (1 - self.beta2 ** (self.i + 1))

        # Perform update.
        if self.local_rates:
            denom = B.sqrt(B.mean(v_corr)) + self.epsilon
        else:
            denom = B.sqrt(v_corr) + self.epsilon
        x -= self.rate * m_corr / denom

        # Increase iteration number.
        self.i += 1

        return x



import numpy as np
import pytest
import torch
from varz import Vars
import lab as B

# noinspection PyUnresolvedReferences
from .util import (
    approx,
    approx,
    KV,
    dtype,
    vs,
    vs_source,
    assert_lower_triangular,
    assert_positive_definite,
    assert_orthogonal,
)


def test_get_latent_vars():
    vs = Vars(np.int)

    # This test also tests that `Vars.get_vars` always returns the collection
    # of variables in the right order. This is important for optimisation.

    # Initialise some variables.
    vs.ubnd(1, name="a")
    vs.ubnd(2, name="1/b")
    vs.ubnd(3, name="2/c")
    vs.unbounded(4, name="2/d")  # Test alias.

    # Test getting all.
    assert vs.get_latent_vars() == [1, 2, 3, 4]
    assert vs.get_latent_vars(return_indices=True) == [0, 1, 2, 3]

    # Test that names must exist.
    with pytest.raises(ValueError):
        vs.get_latent_vars("e")

    # Test some queries.
    assert vs.get_latent_vars("a") == [1]
    assert vs.get_latent_vars("a", "*/b") == [1, 2]
    assert vs.get_latent_vars("*/c", "a") == [1, 3]
    assert vs.get_latent_vars("*/c", "*/b", "a") == [1, 2, 3]

    assert vs.get_latent_vars("a", return_indices=True) == [0]
    assert vs.get_latent_vars("a", "*/b", return_indices=True) == [0, 1]
    assert vs.get_latent_vars("*/c", "a", return_indices=True) == [0, 2]
    assert vs.get_latent_vars("*/c", "*/b", "a", return_indices=True) == [0, 1, 2]

    # Test some more queries.
    assert vs.get_latent_vars("1/*") == [2]
    assert vs.get_latent_vars("2/*") == [3, 4]
    assert vs.get_latent_vars("1/*", "2/*") == [2, 3, 4]

    assert vs.get_latent_vars("1/*", return_indices=True) == [1]
    assert vs.get_latent_vars("2/*", return_indices=True) == [2, 3]
    assert vs.get_latent_vars("1/*", "2/*", return_indices=True) == [1, 2, 3]

    # Test even more queries.
    assert vs.get_latent_vars("*/b", "1/*") == [2]
    assert vs.get_latent_vars("a", "2/*") == [1, 3, 4]
    assert vs.get_latent_vars("a", "2/d", "2/*") == [1, 3, 4]
    assert vs.get_latent_vars("2/d", "2/c", "a", "1/*") == [1, 2, 3, 4]
    assert vs.get_latent_vars("1/*") == [2]
    assert vs.get_latent_vars("2/*") == [3, 4]
    assert vs.get_latent_vars("1/*", "2/*") == [2, 3, 4]

    assert vs.get_latent_vars("*/b", "1/*", return_indices=True) == [1]
    assert vs.get_latent_vars("a", "2/*", return_indices=True) == [0, 2, 3]
    assert vs.get_latent_vars("a", "2/d", "2/*", return_indices=True) == [0, 2, 3]
    assert vs.get_latent_vars("2/d", "2/c", "a", "1/*", return_indices=True) == [
        0,
        1,
        2,
        3,
    ]
    assert vs.get_latent_vars("1/*", return_indices=True) == [1]
    assert vs.get_latent_vars("2/*", return_indices=True) == [2, 3]
    assert vs.get_latent_vars("1/*", "2/*", return_indices=True) == [1, 2, 3]


def test_get_latent_vars_cache_clearing(vs):
    vs.ubnd(name="var_a")
    assert vs.get_latent_vars("var_*", return_indices=True) == [0]
    vs.ubnd(name="var_b")
    assert vs.get_latent_vars("var_*", return_indices=True) == [0, 1]


def test_unbounded(vs_source):
    for _ in range(10):
        vs_source.ubnd()
        vs_source.unbounded()


def test_unbounded_init(vs):
    vs.ubnd(1, name="x")
    approx(vs["x"], 1)

    # Test that explicit data type can be given. This should work the same
    # for all variable getters, so we only test it once.
    assert B.dtype(vs.ubnd(1, name="y", dtype=int)) == np.int64


def test_unbounded_assignment(vs):
    vs.ubnd(1, name="x")
    vs.assign("x", 2)
    approx(vs["x"], 2)


def test_positive(vs_source):
    for _ in range(10):
        assert vs_source.pos() >= 0
        assert vs_source.positive() >= 0


def test_positive_init(vs):
    vs.pos(1, name="x")
    approx(vs["x"], 1)


def test_positive_assignment(vs):
    vs.pos(1, name="x")
    vs.assign("x", 2)
    approx(vs["x"], 2)


def test_bounded(vs_source):
    for _ in range(10):
        assert 10 <= vs_source.bnd(lower=10, upper=11) <= 11
        assert 10 <= vs_source.bounded(lower=10, upper=11) <= 11


def test_bounded_init(vs):
    vs.bnd(2, name="x", lower=1, upper=4)
    approx(vs["x"], 2)


def test_bounded_assignment(vs):
    vs.bnd(2, name="x", lower=1, upper=4)
    vs.assign("x", 3)
    approx(vs["x"], 3)


def test_bounded_monotonic(vs):
    vs.bnd(1, lower=0, upper=10)
    vs.bnd(2, lower=0, upper=10)
    vs.bnd(3, lower=0, upper=10)
    assert vs.vars[0] < vs.vars[1] < vs.vars[2]


def test_lower_triangular(vs_source):
    for _ in range(10):
        assert B.shape(vs_source.tril(shape=(5, 5))) == (5, 5)
        assert_lower_triangular(vs_source.tril(shape=(5, 5)))
        assert_lower_triangular(vs_source.lower_triangular(shape=(5, 5)))


def test_lower_triangular_init(vs):
    x = vs.tril(shape=(5, 5))

    vs.tril(x, name="x")
    approx(vs["x"], x)


@pytest.mark.parametrize("shape", [None, 5, (5,), (5, 6)])
def test_lower_triangular_shape(vs, shape):
    with pytest.raises(ValueError):
        vs.tril(shape=shape)


def test_lower_triangular_assignment(vs):
    x = vs.tril(shape=(5, 5))

    vs.tril(shape=(5, 5), name="x")
    vs.assign("x", x)
    approx(vs["x"], x)


def test_positive_definite(vs_source):
    for _ in range(10):
        assert B.shape(vs_source.pd(shape=(5, 5))) == (5, 5)
        assert_positive_definite(vs_source.pd(shape=(5, 5)))
        assert_positive_definite(vs_source.positive_definite(shape=(5, 5)))


def test_positive_definite_init(vs):
    x = vs.pd(shape=(5, 5))

    vs.pd(x, name="x")
    approx(vs["x"], x)


@pytest.mark.parametrize("shape", [None, 5, (5,), (5, 6)])
def test_positive_definite_shape(vs, shape):
    with pytest.raises(ValueError):
        vs.pd(shape=shape)


def test_positive_definite_assignment(vs):
    x = vs.pd(shape=(5, 5))

    vs.pd(shape=(5, 5), name="x")
    vs.assign("x", x)
    approx(vs["x"], x)


@pytest.mark.parametrize("method", ["svd", "expm", "cayley"])
def test_orthogonal(vs_source, method):
    for i in range(10):
        assert B.shape(vs_source.orth(shape=(5, 5), method=method)) == (5, 5)
        assert_orthogonal(vs_source.orth(shape=(5, 5), method=method))
        assert_orthogonal(vs_source.orthogonal(shape=(5, 5), method=method))


def test_orthogonal_method(vs_source):
    with pytest.raises(ValueError):
        vs_source.orth(shape=(5, 5), method="bla")


@pytest.mark.parametrize("method", ["expm", "cayley"])
def test_orthogonal_init(vs, method):
    x = vs.orth(shape=(5, 5), method=method)

    vs.orth(x, name="x", method=method)
    approx(vs["x"], x)


@pytest.mark.parametrize("shape", [(5, 5), (3, 7), (7, 3)])
def test_orthogonal_init_svd(vs, shape):
    x = vs.orth(shape=shape, method="svd")

    vs.orth(x, name="x", method="svd")
    approx(vs["x"], x)


@pytest.mark.parametrize("method", ["expm", "cayley"])
@pytest.mark.parametrize("shape", [None, 5, (5,), (5, 6)])
def test_orthogonal_shape(vs, shape, method):
    with pytest.raises(ValueError):
        vs.orth(shape=shape, method=method)


@pytest.mark.parametrize("shape", [None, 5, (5,)])
def test_orthogonal_shape_svd(vs, shape):
    with pytest.raises(ValueError):
        vs.orth(shape=shape, method="svd")


@pytest.mark.parametrize("method", ["svd", "expm", "cayley"])
def test_orthogonal_assignment(vs, method):
    x = vs.orth(shape=(5, 5), method=method)

    vs.orth(shape=(5, 5), name="x", method=method)
    vs.assign("x", x)
    approx(vs["x"], x)


def test_init_shape_ambiguity(vs):
    with pytest.raises(ValueError):
        vs.ubnd(np.ones(1), shape=())


def test_init_broadcasting(vs):
    vs.ubnd(2, shape=(2,), name="x")
    approx(vs["x"], 2 * B.ones(2))


def test_init_shape_checking(vs):
    vs.ubnd(np.array([1]), shape=(1,), name="x")
    vs.ubnd(np.array([1]), shape=[1], name="y")


def test_get_set_latent_vector(dtype):
    vs = Vars(dtype=dtype)

    # Test stacking a matrix and a vector.
    vs.ubnd(shape=(2,), name="a", init=np.array([1, 2]))
    vs.ubnd(shape=(2, 2), name="b", init=np.array([[3, 4], [5, 6]]))
    approx(vs.get_latent_vector("a", "b"), [1, 2, 3, 4, 5, 6])

    # Test setting elements.
    vs.set_latent_vector(np.array([6, 5, 4, 3, 2, 1]), "a", "b")
    approx(vs["a"], np.array([6, 5]))
    approx(vs["b"], np.array([[4, 3], [2, 1]]))

    # Test setting elements in a differentiable way. This should allow for
    # any values.
    vs.set_latent_vector(
        np.array(["1", "2", "3", "4", "5", "6"]), "a", "b", differentiable=True
    )
    assert np.all(vs["a"] == ["1", "2"])
    assert np.all(vs["b"] == np.array([["3", "4"], ["5", "6"]]))


def test_copy_torch():
    vs = Vars(torch.float64)

    # Create a variable.
    vs.pos(1, name="a")

    # Make a normal and detached copy.
    vs_copy = vs.copy()
    vs_detached = vs.copy(detach=True)

    # Require gradients for both.
    vs.requires_grad(True)
    vs_detached.requires_grad(True)

    # Do a backward pass.
    (vs_detached["a"] ** 2).backward()

    # Check that values are equal, but gradients only computed for one.
    assert vs["a"] == 1
    assert vs.get_latent_vars("a")[0].grad is None
    assert vs_detached["a"] == 1
    assert vs_detached.get_latent_vars("a")[0].grad == 2

    # Check that copied fields are, in fact, copies.
    # is also copied.
    del vs.transforms[:]
    del vs.inverse_transforms[:]
    del vs.vars[:]
    vs.name_to_index.clear()

    for vs_copied in [vs_copy, vs_detached]:
        assert len(vs_copied.transforms) > 0
        assert len(vs_copied.inverse_transforms) > 0
        assert len(vs_copied.vars) > 0
        assert len(vs_copied.name_to_index) > 0


def test_copy_f():
    vs = Vars(np.float64)
    vs.unbounded(1, name="a")
    vs.unbounded(2, name="b")

    # Make a copy and apply a function that sets everything to zero.
    vs_copy = vs.copy(f=lambda x: 0 * x)

    assert vs["a"] == 1
    assert vs["b"] == 2
    assert vs_copy["a"] == 0
    assert vs_copy["b"] == 0


def test_requires_grad_detach_vars_torch():
    vs = Vars(torch.float64)
    vs.pos(1, name="a")

    # Test that gradients need to first be required.
    with pytest.raises(RuntimeError):
        (2 * vs["a"]).backward()

    # Test that gradients can be required and are then computed.
    vs.requires_grad(True)
    (2 * vs["a"]).backward()
    assert vs.vars[0].grad is not None

    # Test that variables can be detached.
    vs.pos(1, name="b")
    result = 2 * vs["b"]
    vs.detach()
    with pytest.raises(RuntimeError):
        result.backward()


def test_source():
    vs = Vars(np.float32, source=np.ones(10))

    assert vs.ubnd() == 1.0
    approx(vs.pos(shape=(5,)), np.exp(np.ones(5, dtype=np.float32)))
    approx(vs.pos(), np.exp(np.float32(1.0)))
    with pytest.raises(ValueError):
        vs.pos(shape=(5,))

    # Test that the source variables are casted to the right data type.

    vs = Vars(np.float32, source=np.array([1]))
    assert vs.ubnd().dtype == np.float32

    vs = Vars(np.float64, source=np.array([1]))
    assert vs.ubnd().dtype == np.float64


def test_names(vs):
    assert vs.names == []
    vs.ubnd()
    assert vs.names == []
    vs.pos(name="a")
    assert vs.names == ["a"]
    vs.bnd(name="b")
    assert vs.names == ["a", "b"]
    vs.ubnd(name="c")
    assert vs.names == ["a", "b", "c"]


def test_print(vs):
    with KV() as mock:
        vs.print()
        assert mock.keys == []
        assert mock.values == []

        vs.ubnd()
        vs.print()
        assert mock.keys == []
        assert mock.values == []

        vs.print()
        vs.ubnd(1, name="a")
        vs.print()
        assert mock.keys == ["a"]
        assert mock.values == [1]

        vs.ubnd(2, name="b")
        vs.print()
        assert mock.keys == ["a", "a", "b"]
        assert mock.values == [1, 1, 2]

        vs.ubnd(3, name="c")
        vs.print()
        assert mock.keys == ["a", "a", "b", "a", "b", "c"]
        assert mock.values == [1, 1, 2, 1, 2, 3]



import pytest

from varz import (
    sequential,
    namespace,
    Unbounded,
    Positive,
    Bounded,
    LowerTriangular,
    PositiveDefinite,
    Orthogonal,
    parametrised,
)
from varz.spec import _extract_prefix_and_f

# noinspection PyUnresolvedReferences
from .util import (
    vs,
    assert_lower_triangular,
    assert_positive_definite,
    assert_orthogonal,
)


def test_extract_prefix_and_f():
    def f():
        pass

    assert _extract_prefix_and_f(f, default="var") == ("var", f)
    assert _extract_prefix_and_f("", default="var") == ("", None)
    assert _extract_prefix_and_f("test", default="var") == ("test", None)


@pytest.mark.parametrize(
    "decorator, names",
    [
        (sequential, ["z", "var0", "var1", "var2"]),
        (sequential(), ["z", "var0", "var1", "var2"]),
        (sequential("x"), ["z", "x0", "x1", "x2"]),
    ],
)
def test_sequential(vs, decorator, names):
    vs.ubnd(0, name="z")

    @decorator
    def f(x, vs_, y):
        assert x == 1
        assert y == 2
        return vs_["z"], vs_.ubnd(), vs_.pos(), vs_.bnd(lower=10, upper=11)

    # Test that the same variables are retrieved.
    assert f(1, vs, 2) == f(1, vs, 2)
    assert vs.names == names
    assert f(1, vs, 2) == f(1, vs, 2)
    assert vs.names == names

    # Test correctness of variables.
    assert f(1, vs, 2)[0] == 0
    assert f(1, vs, 2)[2] >= 0
    assert 10 <= f(1, vs, 2)[3] <= 11


def test_sequential_unbounded(vs):
    @sequential
    def f(vs_):
        return vs_.ubnd(1)

    assert f(vs) == 1


def test_sequential_positive(vs):
    @sequential
    def f(vs_):
        return vs_.pos()

    assert f(vs) > 0


def test_sequential_bounded(vs):
    @sequential
    def f(vs_):
        return vs_.bnd(lower=10, upper=11)

    assert 10 <= f(vs) <= 11


def test_sequential_lower_triangular(vs):
    @sequential
    def f(vs_):
        return vs_.tril(shape=(5, 5))

    assert_lower_triangular(f(vs))


def test_sequential_positive_definite(vs):
    @sequential
    def f(vs_):
        return vs_.pd(shape=(5, 5))

    assert_positive_definite(f(vs))


def test_sequential_orthogonal(vs):
    @sequential
    def f(vs_):
        return vs_.orth(shape=(5, 5))

    assert_orthogonal(f(vs))


def test_namespace(vs):
    @namespace("test")
    def f(vs_):
        return [
            vs_.ubnd(name="a"),
            vs_.ubnd(),
            vs_.ubnd(name="b"),
        ]

    f(vs)
    names = ["test.a", "test.b"]

    # Check that the same variables are retrieved.
    assert f(vs)[0] == f(vs)[0] and f(vs)[2] == f(vs)[2]
    assert vs.names == names
    assert f(vs)[0] == f(vs)[0] and f(vs)[2] == f(vs)[2]
    assert vs.names == names


def test_namespace_sequential(vs):
    @namespace("test")
    @sequential
    def f(vs_):
        vs_.ubnd(name="a")
        vs_.ubnd()
        vs_.ubnd(name="b")

    f(vs)
    names = ["test.a", "test.var0", "test.b"]

    # Check that the same variables are retrieved.
    assert f(vs) == f(vs)
    assert vs.names == names
    assert f(vs) == f(vs)
    assert vs.names == names


def test_struct(vs):
    def f(vs_):
        params = vs_.struct
        res = []

        res.append(params.x1.positive())
        res.append(params.x1.positive(name="var1"))
        res.append(params.y1.positive())
        res.append(params.y1.positive(name="var2"))
        res.append(params["key1"].positive())
        res.append(params["key1"].positive(name="var3"))

        for pi, i in zip(params, [4, 5, 6]):
            res.append(pi.bounded())
            res.append(pi.bounded(name=f"var{i}"))

        params = params.level

        res.append(params.x2.positive())
        res.append(params.x2.positive(name="var7"))
        res.append(params.y2.positive())
        res.append(params.y2.positive(name="var8"))
        res.append(params["key2"].positive())
        res.append(params["key2"].positive(name="var9"))

        for pi, i in zip(params, [10, 11, 12]):
            res.append(pi.bounded())
            res.append(pi.bounded(name=f"var{i}"))

        return res

    f(vs)
    names = [
        "x1",
        "x1.var1",
        "y1",
        "y1.var2",
        "[key1]",
        "[key1].var3",
        "[0]",
        "[0].var4",
        "[1]",
        "[1].var5",
        "[2]",
        "[2].var6",
        "level.x2",
        "level.x2.var7",
        "level.y2",
        "level.y2.var8",
        "level[key2]",
        "level[key2].var9",
        "level[0]",
        "level[0].var10",
        "level[1]",
        "level[1].var11",
        "level[2]",
        "level[2].var12",
    ]

    # Check that the same variables are retrieved.
    assert f(vs) == f(vs)
    assert vs.names == names
    assert f(vs) == f(vs)
    assert vs.names == names

    # Check extraction by calling.
    assert vs.struct.x1.positive() == vs.struct.x1()

    # Test getting a regex to match everything in a path.
    assert vs.struct.level[0].all() == "level[0].*"

    # Test negative indices.
    assert vs.struct.level[-1]._path == "level[2]"
    assert vs.struct.level[-2]._path == "level[1]"

    # Test getting the next one in a numbered array.
    assert next(vs.struct.level)._path == "level[3]"
    # Check that `level[3]` does not have to be a variable itself for `level[3]` to
    # be detected.
    vs.struct.level[3].some_variable.positive()
    assert "level[3]" not in vs.names
    assert next(vs.struct.level)._path == "level[4]"

    # Test length.
    assert len(vs.struct.level) == 4
    assert len(vs.struct.level_does_not_exist) == 0

    # Test going up one level.
    assert vs.struct.a.b.c.up()._path == "a.b"
    assert vs.struct.a.b.c.up("c")._path == "a.b"
    with pytest.raises(AssertionError):
        vs.struct.a.b.c.up("d")


@pytest.mark.parametrize(
    "decorator, types, names",
    [
        (
            parametrised,
            [Unbounded, Positive, Bounded(lower=10, upper=11)],
            ["w", "x", "y", "z"],
        ),
        (
            parametrised(),
            [Unbounded, Positive, Bounded(lower=10, upper=11)],
            ["w", "x", "y", "z"],
        ),
        (
            parametrised("var_"),
            [Unbounded, Positive, Bounded(lower=10, upper=11)],
            ["w", "var_x", "var_y", "var_z"],
        ),
    ],
)
def test_parametrised(vs, decorator, types, names):
    vs.ubnd(0, name="w")

    @decorator
    def f(a, x: types[0], vs_, y: types[1], b, z: types[2], c=None):
        assert a == 1
        assert b == 2
        assert c is None
        return vs_["w"], x, y, z

    # Test that the same variables are retrieved
    assert f(1, vs, 2) == f(1, vs, 2)
    assert vs.names == names
    assert f(1, vs, 2) == f(1, vs, 2)
    assert vs.names == names

    # Test correctness of variables.
    assert f(1, vs, 2)[0] == 0
    assert f(1, vs, 2)[2] >= 0
    assert 10 <= f(1, vs, 2)[3] <= 11


def test_parametrised_arguments(vs):
    @parametrised
    def f(vs_, x: Unbounded, a, y: Unbounded, b=2, z: Unbounded = 10, c=3):
        return a, b, c

    with pytest.raises(ValueError):
        f(vs)

    assert f(vs, 1) == (1, 2, 3)
    assert f(vs, a=1) == (1, 2, 3)

    assert f(vs, 1, 4) == (1, 4, 3)
    assert f(vs, 1, b=4) == (1, 4, 3)
    assert f(vs, a=1, b=4) == (1, 4, 3)

    assert f(vs, 1, 4, 5) == (1, 4, 5)
    assert f(vs, 1, 4, c=5) == (1, 4, 5)
    assert f(vs, a=1, b=4, c=5) == (1, 4, 5)

    # Test unspecified positional argument.
    with pytest.raises(ValueError):
        f(vs, 1, 2, 3, 4)

    # Test unspecified keyword argument.
    with pytest.raises(ValueError):
        f(vs, 1, 2, d=3)


def test_parametrised_variable_container(vs):
    @parametrised
    def f(x, y, z: Unbounded):
        pass

    # Test that a variable container must be present.
    with pytest.raises(ValueError):
        f(1, 2)

    # Test that only one variable container can be present.
    with pytest.raises(ValueError):
        f(vs, vs)


def test_parametrised_double_initial_value(vs):
    @parametrised
    def f(vs_, x: Unbounded(init=5) = 5):
        pass

    with pytest.raises(ValueError):
        f(vs)


def test_parametrised_double_name(vs):
    @parametrised
    def f(vs_, x: Unbounded(name="x")):
        pass

    with pytest.raises(ValueError):
        f(vs)


def test_parametrised_unbounded(vs):
    @parametrised
    def f(vs_, x: Unbounded = 1):
        return x

    assert f(vs) == 1


def test_parametrised_positive(vs):
    @parametrised
    def f(vs_, x: Positive):
        return x

    assert f(vs) > 0


def test_parametrised_bounded(vs):
    @parametrised
    def f(vs_, x: Bounded(lower=10, upper=11)):
        return x

    assert 10 <= f(vs) <= 11


def test_parametrised_lower_triangular(vs):
    @parametrised
    def f(vs_, x: LowerTriangular(shape=(5, 5))):
        return x

    assert_lower_triangular(f(vs))


def test_parametrised_positive_definite(vs):
    @parametrised
    def f(vs_, x: PositiveDefinite(shape=(5, 5))):
        return x

    assert_positive_definite(f(vs))


def test_parametrised_orthogonal(vs):
    @parametrised
    def f(vs_, x: Orthogonal(shape=(5, 5))):
        return x

    assert_orthogonal(f(vs))



import lab as B
import pytest

from varz import pack, unpack
from .util import approx


def test_pack_unpack():
    a, b, c = B.randn(5, 10), B.randn(20), B.randn(5, 1, 15)

    # Test packing.
    package = pack(a, b, c)
    assert B.rank(package) == 1

    # Test unpacking.
    a2, b2, c2 = unpack(package, B.shape(a), B.shape(b), B.shape(c))
    approx(a, a2)
    approx(b, b2)
    approx(c, c2)

    # Check that the package must be a vector.
    with pytest.raises(ValueError):
        unpack(B.randn(2, 2), (2, 2))



import logging
import warnings
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from functools import reduce
from operator import mul

import lab as B
import numpy as np
import wbml.out
from lab.shape import Shape
from plum import Dispatcher

from .util import (
    pack,
    unpack,
    match,
    lazy_tf as tf,
    lazy_torch as torch,
    lazy_jnp as jnp,
)

__all__ = ["Provider", "Vars"]

log = logging.getLogger(__name__)

_dispatch = Dispatcher()


@_dispatch
def _assign(x: B.NPNumeric, value: B.Numeric):
    np.copyto(x, value)
    return x


@_dispatch
def _assign(x: B.TFNumeric, value: B.Numeric):
    # The return value of `x.assign` tends to behave unpredictably. We therefore just
    # return `x`.
    x.assign(value)
    return x


@_dispatch
def _assign(x: B.TorchNumeric, value: B.Numeric):
    if not isinstance(value, B.TorchNumeric):
        value = torch.tensor(value, dtype=x.dtype, device=x.device)
    x.data.copy_(value)
    return x


@_dispatch
def _assign(x: B.JAXNumeric, value: B.Numeric):
    return B.to_active_device(jnp.array(value, dtype=x.dtype))


class Provider(metaclass=ABCMeta):
    @abstractmethod
    def unbounded(
        self, init=None, shape=None, dtype=None, name=None
    ):  # pragma: no cover
        """Get an unbounded variable.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (tuple[int], optional): Shape of the variable. Defaults to scalar.
            dtype (data type, optional): Data type of the variable. Defaults to that
                of the storage.
            name (str, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def ubnd(self, *args, **kw_args):
        """Alias for :meth:`.vars.Provider.unbounded`."""
        return self.unbounded(*args, **kw_args)

    def get(self, *args, **kw_args):  # pragma: no cover
        warnings.warn(
            "The method `get` is deprecated. Please use `unbounded` or `ubnd` instead.",
            category=DeprecationWarning,
        )
        return self.unbounded(*args, **kw_args)

    @abstractmethod
    def positive(
        self, init=None, shape=None, dtype=None, name=None
    ):  # pragma: no cover
        """Get a positive variable.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (tuple[int], optional): Shape of the variable. Defaults to scalar.
            dtype (data type, optional): Data type of the variable. Defaults to that
                of the storage.
            name (str, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def pos(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.positive`."""
        return self.positive(*args, **kw_args)

    @abstractmethod
    def bounded(
        self, init=None, lower=1e-4, upper=1e4, shape=None, dtype=None, name=None
    ):  # pragma: no cover
        """Get a bounded variable.

        Args:
            init (tensor, optional): Initialisation of the variable.
            lower (tensor, optional): Lower bound. Defaults to `1e-4`.
            upper (tensor, optional): Upper bound. Defaults to `1e4`.
            shape (tuple[int], optional): Shape of the variable. Defaults to scalar.
            dtype (data type, optional): Data type of the variable. Defaults to that
                of the storage.
            name (hashable, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def bnd(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.bounded`."""
        return self.bounded(*args, **kw_args)

    @abstractmethod
    def lower_triangular(
        self, init=None, shape=None, dtype=None, name=None
    ):  # pragma: no cover
        """Get a lower-triangular matrix.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (int, optional): Number of rows and columns of the matrix.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (hashable, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def tril(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.lower_triangular`."""
        return self.lower_triangular(*args, **kw_args)

    @abstractmethod
    def positive_definite(
        self, init=None, shape=None, dtype=None, name=None
    ):  # pragma: no cover
        """Get a positive-definite matrix.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (int, optional): Number of rows and columns of the matrix.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (hashable, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def pd(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.positive_definite`."""
        return self.positive_definite(*args, **kw_args)

    @abstractmethod
    def orthogonal(
        self, init=None, shape=None, dtype=None, name=None, method="svd"
    ):  # pragma: no cover
        """Get an orthogonal matrix.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (int, optional): Number of rows and columns of the matrix.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (hashable, optional): Name of the variable.
            method ('svd', 'expm' or 'cayley'): Parametrisation. Method of
                parametrisation. Defaults to 'svd'.

        Returns:
            tensor: Variable.
        """

    def orth(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.orthogonal`."""
        return self.orthogonal(*args, **kw_args)

    @abstractmethod
    def __getitem__(self, name):  # pragma: no cover
        """Get a variable by name.

        Args:
            name (hashable): Name of variable.

        Returns:
            tensor: Variable.
        """

    @property
    def struct(self):
        """:class:`.spec.Structlike`: Name variables by indexing into the variable
        container like a struct."""
        # We perform the import here to avoid a circular import.
        from .spec import Struct

        return Struct(self)


@_dispatch
def _check_matrix_shape(shape, square=True):
    raise ValueError(f"Object {shape} is not a shape.")


@_dispatch
def _check_matrix_shape(shape: tuple, square=True):
    if len(shape) != 2:
        raise ValueError(f"Shape {shape} must be the shape of a matrix.")
    if square and shape[0] != shape[1]:
        raise ValueError(f"Shape {shape} must be square.")


def _check_init_shape(init, shape):
    if init is None and shape is None:
        raise ValueError(
            f"The shape must be given to automatically initialise "
            f"a matrix variable."
        )
    if shape is None:
        shape = B.shape(init)
    return init, shape


class Vars(Provider):
    """Variable storage.

    Args:
        dtype (data type): Data type of the variables.
        source (tensor, optional): Tensor to source variables from. Defaults to
            not being used.
    """

    def __init__(self, dtype, source=None):
        self.dtype = dtype

        # Source:
        self.source = source
        self.source_index = 0

        # Storage:
        self.vars = []
        self.transforms = []
        self.inverse_transforms = []

        # Lookup:
        self.name_to_index = OrderedDict()
        self._get_latent_vars_cache = {}

    def _resolve_dtype(self, dtype):
        if dtype is None:
            return self.dtype
        else:
            return dtype

    def _get_var(
        self,
        transform,
        inverse_transform,
        init,
        generate_init,
        shape,
        shape_latent,
        dtype,
        name,
    ):
        # If the name already exists, return that variable.
        try:
            return self[name]
        except KeyError:
            pass

        # A new variable will be added. Clear lookup cache.
        self._get_latent_vars_cache.clear()

        # Resolve data type.
        dtype = self._resolve_dtype(dtype)

        # If no source is provided, get the latent from from the provided
        # initialiser.
        if self.source is None:
            # Resolve initialisation.
            if init is None:
                init = generate_init(shape=shape, dtype=dtype)
            else:
                init = B.cast(dtype, init)

            # Ensure that the initialisation is on the right device.
            init = B.to_active_device(init)

            # Allow broadcasting in the initialisation.
            if shape is not None:
                init = init * B.ones(B.dtype(init), *shape)

            # Double check the shape of the initialisation.
            if shape is not None and Shape(*shape) != Shape(*B.shape(init)):
                raise ValueError(
                    f"Shape of initial value {B.shape(init)} is not equal to the "
                    f"desired shape {shape}."
                )

            # Construct optimisable variable.
            latent = inverse_transform(init)
            if isinstance(self.dtype, B.TFDType):
                latent = tf.Variable(latent)
            elif isinstance(self.dtype, B.TorchDType):
                pass  # All is good in this case.
            elif isinstance(self.dtype, B.JAXDType):
                latent = jnp.array(latent)
            else:
                # Must be a NumPy data type.
                assert isinstance(self.dtype, B.NPDType)
                latent = np.array(latent)
        else:
            # Get the latent variable from the source.
            length = reduce(mul, shape_latent, 1)
            latent_flat = self.source[self.source_index : self.source_index + length]
            self.source_index += length

            # Cast to the right data type.
            latent = B.cast(dtype, B.reshape(latent_flat, *shape_latent))

        # Store transforms.
        self.vars.append(latent)
        self.transforms.append(transform)
        self.inverse_transforms.append(inverse_transform)

        # Get index of the variable.
        index = len(self.vars) - 1

        # Store name if given.
        if name is not None:
            self.name_to_index[name] = index

        # Generate the variable and return.
        return transform(latent)

    def unbounded(self, init=None, shape=None, dtype=None, name=None):
        # If nothing is specific, generate a scalar.
        if init is None and shape is None:
            shape = ()

        def generate_init(shape, dtype):
            return B.randn(dtype, *shape)

        return self._get_var(
            transform=lambda x: x,
            inverse_transform=lambda x: x,
            init=init,
            generate_init=generate_init,
            shape=shape,
            shape_latent=shape,
            dtype=dtype,
            name=name,
        )

    def positive(self, init=None, shape=None, dtype=None, name=None):
        # If nothing is specific, generate a scalar.
        if init is None and shape is None:
            shape = ()

        def generate_init(shape, dtype):
            return B.rand(dtype, *shape)

        return self._get_var(
            transform=lambda x: B.exp(x),
            inverse_transform=lambda x: B.log(x),
            init=init,
            generate_init=generate_init,
            shape=shape,
            shape_latent=shape,
            dtype=dtype,
            name=name,
        )

    def bounded(
        self, init=None, lower=1e-4, upper=1e4, shape=None, dtype=None, name=None
    ):
        # If nothing is specific, generate a scalar.
        if init is None and shape is None:
            shape = ()

        def transform(x):
            return lower + (upper - lower) / (1 + B.exp(-x))

        def inverse_transform(x):
            return B.log(x - lower) - B.log(upper - x)

        def generate_init(shape, dtype):
            return lower + B.rand(dtype, *shape) * (upper - lower)

        return self._get_var(
            transform=transform,
            inverse_transform=inverse_transform,
            init=init,
            generate_init=generate_init,
            shape=shape,
            shape_latent=shape,
            dtype=dtype,
            name=name,
        )

    def lower_triangular(self, init=None, shape=None, dtype=None, name=None):
        init, shape = _check_init_shape(init, shape)
        _check_matrix_shape(shape)

        # Result must be square. Get a side.
        side = shape[0]

        def transform(x):
            return B.vec_to_tril(x)

        def inverse_transform(x):
            return B.tril_to_vec(x)

        def generate_init(shape, dtype):
            mat = B.randn(dtype, *shape)
            return transform(B.tril_to_vec(mat))

        shape_latent = (int(side * (side + 1) / 2),)
        return self._get_var(
            transform=transform,
            inverse_transform=inverse_transform,
            init=init,
            generate_init=generate_init,
            shape=shape,
            shape_latent=shape_latent,
            dtype=dtype,
            name=name,
        )

    def positive_definite(self, init=None, shape=None, dtype=None, name=None):
        init, shape = _check_init_shape(init, shape)
        _check_matrix_shape(shape)

        # Result must be square. Get a side.
        side = shape[0]

        def transform(x):
            log_diag = x[:side]
            chol = B.vec_to_tril(x[side:], offset=-1) + B.diag(B.exp(log_diag))
            return B.matmul(chol, chol, tr_b=True)

        def inverse_transform(x):
            chol = B.cholesky(B.reg(x))
            return B.concat(B.log(B.diag(chol)), B.tril_to_vec(chol, offset=-1))

        def generate_init(shape, dtype):
            mat = B.randn(dtype, *shape)
            return B.matmul(mat, mat, tr_b=True)

        shape_latent = (int(side * (side + 1) / 2),)
        return self._get_var(
            transform=transform,
            inverse_transform=inverse_transform,
            init=init,
            generate_init=generate_init,
            shape=shape,
            shape_latent=shape_latent,
            dtype=dtype,
            name=name,
        )

    def orthogonal(self, init=None, shape=None, dtype=None, name=None, method="svd"):
        init, shape = _check_init_shape(init, shape)

        if method == "svd":
            _check_matrix_shape(shape, square=False)
            n, m = shape
            shape_latent = (n, m)

            # Fix singular values.
            sing_vals = B.linspace(self._resolve_dtype(dtype), 1, 2, min(n, m))

            def transform(x):
                u, s, v = B.svd(x)
                # u * v' is the closest orthogonal matrix to x in Frobenius norm.
                return B.matmul(u, v, tr_b=True)

            def inverse_transform(x):
                if n >= m:
                    return x * sing_vals[None, :]
                else:
                    return x * sing_vals[:, None]

            def generate_init(shape, dtype):
                mat = B.randn(dtype, *shape)
                return transform(mat)

        elif method == "expm":
            _check_matrix_shape(shape)
            side = shape[0]
            shape_latent = (int(side * (side + 1) / 2 - side),)

            def transform(x):
                tril = B.vec_to_tril(x, offset=-1)
                skew = tril - B.transpose(tril)
                return B.expm(skew)

            def inverse_transform(x):
                return B.tril_to_vec(B.logm(x), offset=-1)

            def generate_init(shape, dtype):
                mat = B.randn(dtype, *shape)
                return transform(B.tril_to_vec(mat, offset=-1))

        elif method == "cayley":
            _check_matrix_shape(shape)
            side = shape[0]
            shape_latent = (int(side * (side + 1) / 2 - side),)

            def transform(x):
                tril = B.vec_to_tril(x, offset=-1)
                skew = tril - B.transpose(tril)
                eye = B.eye(skew)
                return B.solve(eye + skew, eye - skew)

            def inverse_transform(x):
                eye = B.eye(x)
                skew = B.solve(eye + x, eye - x)
                return B.tril_to_vec(skew, offset=-1)

            def generate_init(shape, dtype):
                mat = B.randn(dtype, *shape)
                return transform(B.tril_to_vec(mat, offset=-1))

        else:
            raise ValueError(f'Unknown parametrisation "{method}".')

        return self._get_var(
            transform=transform,
            inverse_transform=inverse_transform,
            init=init,
            generate_init=generate_init,
            shape=shape,
            shape_latent=shape_latent,
            dtype=dtype,
            name=name,
        )

    def __getitem__(self, name):
        index = self.name_to_index[name]
        return self.transforms[index](self.vars[index])

    def assign(self, name, value):
        """Assign a value to a variable.

        Args:
            name (hashable): Name of variable to assign value to.
            value (tensor): Value to assign.

        Returns:
            tensor: Assignment result.
        """
        index = self.name_to_index[name]
        # Overwrite data.
        self.vars[index] = _assign(
            self.vars[index], self.inverse_transforms[index](value)
        )
        return self.vars[index]

    def copy(self, detach=False, f=lambda x: x):
        """Create a copy of the variable manager that shares the variables.

        Args:
            detach (bool, optional): Detach the variables in PyTorch. Defaults
                to `False`.
            f (function, optional): Apply this function to every latent variable.

        Returns:
            :class:`.vars.Vars`: Copy.
        """
        vs = Vars(dtype=self.dtype)
        vs.transforms = list(self.transforms)
        vs.inverse_transforms = list(self.inverse_transforms)
        vs.name_to_index = OrderedDict(self.name_to_index)
        vs.vars = [f(x) for x in self.vars]
        if detach:
            vs.detach()
        return vs

    def detach(self):
        """Detach all variables held in PyTorch."""
        self.vars = [v.detach() for v in self.vars]

    def requires_grad(self, value, *names):
        """Set which variables require a gradient in PyTorch.

        Args:
            value (bool): Require a gradient.
            *names (hashable): Specify variables by name.
        """
        for var in self.get_latent_vars(*names):
            var.requires_grad_(value)

    def get_latent_vars(self, *names, return_indices=False):
        """Get latent variables.

        If no arguments are supplied, then all latent variables are retrieved.
        Furthermore, the same collection of variables is guaranteed to be
        returned in the same order.

        Args:
            *names (hashable): Get variables by name.
            return_indices (bool, optional): Get the indices of the variables instead.
                Defaults to `False`.

        Returns:
            list: Matched latent variables or their indices, depending on the
                value of `indices`.
        """
        # If nothing is specified, return all latent variables.
        if len(names) == 0:
            if return_indices:
                return list(range(len(self.vars)))
            else:
                return self.vars

        # Attempt to use cache.
        try:
            indices = self._get_latent_vars_cache[names]
        except KeyError:
            # Collect indices of matches.
            indices = set()
            for name in names:
                a_match = False
                for k, v in self.name_to_index.items():
                    if match(name, k):
                        indices |= {v}
                        a_match = True

                # Check that there was a match.
                if not a_match:
                    raise ValueError(f'No variable matching "{name}".')

            # Sort the indices for a consistent result.
            indices = sorted(indices)

            # Store in cache before proceeding.
            self._get_latent_vars_cache[names] = indices

        # Return indices if asked for. Otherwise, return variables.
        if return_indices:
            return indices
        else:
            return [self.vars[i] for i in indices]

    def get_vars(self, *args, **kw_args):  # pragma: no cover
        warnings.warn(
            "The method `get_vars` is deprecated. Please use `get_latent_vars` "
            "instead.",
            category=DeprecationWarning,
        )
        return self.get_latent_vars(*args, **kw_args)

    def get_latent_vector(self, *names):
        """Get all the latent variables stacked in a vector.

        If no arguments are supplied, then all latent variables are retrieved.

        Args:
            *names (hashable): Get variables by name.

        Returns:
            tensor: Vector consisting of all latent values
        """
        return pack(*self.get_latent_vars(*names))

    def get_vector(self, *args, **kw_args):  # pragma: no cover
        warnings.warn(
            "The method `get_vector` is deprecated. Please use `get_latent_vector` "
            "instead.",
            category=DeprecationWarning,
        )
        return self.get_latent_vector(*args, **kw_args)

    def set_latent_vector(self, values, *names, differentiable=False):
        """Set all the latent variables by values from a vector.

        If no arguments are supplied, then all latent variables are retrieved.

        Args:
            values (tensor): Vector to set the variables to.
            *names (hashable): Set variables by name.
            differentiable (bool, optional): Differentiable assignment. Defaults
                to `False`.

        Returns:
            list: Assignment results.
        """
        values = unpack(values, *map(B.shape, self.get_latent_vars(*names)))

        if differentiable:
            # Do a differentiable assignment.
            for index, value in zip(
                self.get_latent_vars(*names, return_indices=True), values
            ):
                self.vars[index] = value
            return values
        else:
            # Overwrite data.
            assignments = []
            for index, value in zip(
                self.get_latent_vars(*names, return_indices=True), values
            ):
                self.vars[index] = _assign(self.vars[index], value)
                assignments.append(self.vars[index])
            return assignments

    def set_vector(self, *args, **kw_args):  # pragma: no cover
        warnings.warn(
            "The method `set_vector` is deprecated. Please use `set_latent_vector` "
            "instead.",
            category=DeprecationWarning,
        )
        return self.get_latent_vector(*args, **kw_args)

    @property
    def names(self):
        """All available names."""
        return list(self.name_to_index.keys())

    def print(self):
        """Print all variables."""
        for name in self.names:
            wbml.out.kv(name, self[name])
